package btc.loader

import java.io. {

  BufferedReader,
  BufferedWriter,
  File,
  FileOutputStream,
  FileWriter,
  InputStream,
  InputStreamReader,
  IOException,
  OutputStreamWriter

}

import java.net.URL
import java.text.SimpleDateFormat
import java.util.zip.GZIPOutputStream

import net.liftweb.common. { Loggable, Logger }
import net.liftweb.json. { DefaultFormats, parse }

import org.joda.time.LocalDate

import scala.util.control.Breaks. { break, breakable }

class Loader(
  zipped: Boolean,
  requestMax: Int,
  requestDelay: Int,
  from: Long,
  till: Long,
  output: File,
  symbols: Array[String]) extends Logger with Loggable {

  implicit protected val formats = DefaultFormats

  protected val dateFormatter = new SimpleDateFormat("yyyy-MM-dd")

  try {

    val fromUnixtime = from / 1000
    object continue extends Throwable

    // Iterate all markets
    for(value <- parse(request(
      "http://bitcoincharts.com/t/markets.json",
      requestMax,
      requestDelay)).children)
    try {

      val symbol = (value \ "symbol").extract[String]

      if(symbols.isEmpty || symbols.contains(symbol)) {

        var stream: InputStream = null

        try {

          var open = false
          var requestCount = 0

          do {

            try {

              stream = new URL("http://bitcoincharts.com/t/trades.csv?symbol=" +
                symbol + "&start=" + fromUnixtime).openStream
              open = true

            } catch {

              case e: IOException =>
                error("Publishing stopped due to: " + e)

                if(requestMax > 0)
                  requestCount = requestCount + 1

                if(requestCount == requestMax)
                  throw new MaxRequestsExcessedException

                info("Try request in " + requestDelay + "s");

                // Request delay.
                Thread.sleep(requestDelay * 1000)

            }
          } while(!open)
        } catch {

          case e: InterruptedException => sys.exit

        }

        val reader = new BufferedReader(new InputStreamReader(stream))

        val extension =
          if(zipped)
            ".gz"
          else
            ""

        val file = new File(
          output + "/" + symbol + "_trade_" + dateFormatter.format(from) +
          "_" + dateFormatter.format(till) + ".csv" + extension)

        if(file.exists)
          file.delete

        var writer: BufferedWriter = null

        if(zipped)
          writer = new BufferedWriter(new OutputStreamWriter(
            new GZIPOutputStream(new FileOutputStream(file))))
        else
          writer = new BufferedWriter(new FileWriter(file))

        var line = reader.readLine

        writer.write("unixtime,price,amount")
        writer.newLine

        var _from: Long = 0
        var _till: Long = 0
        var lastDate: LocalDate = null
        var numberOfLinesOnDay = 0L
        var numberOfLinesWritten = 0L

        breakable {
          while(line != null) {

            val snapshot = line.split(",")

            if(
              snapshot.size == 3 &&
              snapshot(0).toLong >= fromUnixtime) {

              var newDay = false
              val time = snapshot(0).toLong * 1000

              if(time >= till)
                break

              if(lastDate == null) {

                _from = time
                newDay = true

              } else {

                val today = new LocalDate(time)

                if(lastDate.getDayOfMonth != today.getDayOfMonth) {

                  if(numberOfLinesOnDay > 0)
                    info(symbol + ": EOD " +
                      lastDate.toString + " passed, +" +
                      numberOfLinesOnDay + " lines")

                  newDay = true
                  numberOfLinesOnDay = 0

                }
              }

              if(newDay)
                lastDate = new LocalDate(time)

              _till = time
              writer.write(line)
              writer.newLine

              numberOfLinesOnDay += 1
              numberOfLinesWritten += 1

            }

            line = reader.readLine

          }
        }

        writer.close
        reader.close

        if(numberOfLinesWritten == 0) {

          if(file.exists)
            file.delete

          warn(symbol + ": Skipped because no data available")

        } else {

          _till += 1000 * 60 * 60 * 24

          if(from != _from || till != till) {

            val _file = new File(output + "/" + symbol + "_trade_" +
              dateFormatter.format(_from) + "_" + dateFormatter.format(_till)
              + ".csv" + extension)

            file.renameTo(_file)

          }

          info(symbol + ": All data written succesfully, " + "number of lines " +
            "written: " + numberOfLinesWritten)

        }
  
      }
    } catch {
      case continue =>
    }
  } catch {

    case e: MaxRequestsExcessedException =>
      error("Request stopped due to: Max. attempts reached")

  }

  /** Request feed from url.
   *
   * @param url the url to request
   * @param requestMax the maximum request attepmts
   * @param requestDelay the sleep time till next request
   *
   * @return requested content or null
  */
  protected def request(
    url: String,
    requestMax: Int = 3,
    requestDelay: Int = 5000): String = {

    try {

      var stream: InputStream = null
      var open = false
      var requestCount = 0

      do {

        try {

          stream = new URL(url).openStream
          open = true

        } catch {

          case e: IOException =>
            error("Publishing stopped due to: " + e)

            if(requestMax > 0)
              requestCount = requestCount + 1

            if(requestCount == requestMax)
              throw new MaxRequestsExcessedException

            info("Try request in " + requestDelay + "s");

            // Request delay.
            Thread.sleep(requestDelay * 1000)

        }
      } while(!open)

      val reader = new BufferedReader(new InputStreamReader(stream))

      var content = ""
      var line = reader.readLine

      while(line != null) {

        content += line
        line = reader.readLine

        if(line != null)
          content += "\n"

      }

      reader.close

      return content

    } catch {

      case e: InterruptedException =>

    }

    null

  }
}

